    MODULE chainlinkmod
! .. Use Statements ..
      USE set_precision, ONLY: wp
! ..
    CONTAINS

! Corrected and restructured version in Fortran 90 of
! the original Algol 60 routines previously published as
!
! The argument lists of inlist and outlist could be simplified
! further if arguments like m and n were retrieved from the
! array sizes of the other arguments.
!
! ALGORITHM 100
! ADD ITEM TO CHAIN-LINKED LIST
! and
! ALGORITHM 101
! REMOVE ITEM FROM CHAIN-LINKED LIST
! by
! PHILIP J. KIVIAT
! United States Steel Corp., Appl. Research Lab., Monroeville,
! Penn. ;
      SUBROUTINE inlist(t, info, m, list, n, first, flag, addr, listfull)
! .. Scalar Arguments ..
        INTEGER, INTENT (INOUT) :: first
        INTEGER, INTENT (IN) :: flag, m, n, t
        LOGICAL, INTENT (INOUT) :: listfull
! ..
! .. Array Arguments ..
        INTEGER, INTENT (INOUT) :: addr(:), list(:, :)
        INTEGER, INTENT (IN) :: info(:)
! ..
! .. Local Scalars ..
        INTEGER :: i, j, k, link1, link2
! ..
! inlist adds the information pair {t,info} to the chain-
! link structured matrix list (i,j), where t is an order key >= 0, and
! info(k) an information vector associated with t. info(k) has
! dimension m, list(i,j) has dimensions (n, (m+3)). flag denotes
! the head and tail of list(i,j), and first contains the address of
! the first (lowest order) entry in list(i,j). addr(k) is a vector
! containing the addresses of available (empty) rows in list(i,j).
! Initialization: list(i,m+2) = flag, for some i <= n. If list(i,j) is
! filled the logical variable listfull returns TRUE.
        listfull = .FALSE.
        IF (addr(1)==0) THEN
          listfull = .TRUE.
        ELSE IF (addr(n)/=0) THEN
! Insertion into an empty list.
          i = flag
          link1 = m + 3
          link2 = m + 2
          CALL addlink
          CALL adddata

        ELSE

! There is at least one element to compare against.
          i = first
loop:     DO

            IF (list(i,1)<=t) THEN
! Insert after first item
              link1 = m + 2
              link2 = m + 3
            ELSE
! Insert before first item
              link1 = m + 3
              link2 = m + 2
            END IF

            IF (list(i,link2)/=flag) THEN
! Check the next element for the insertion point.
              k = i
              i = list(i, link2)
! Insertion point found.
! Continue the search.
              IF (list(i,1)>t) THEN
                j = addr(1)
                list(j, link1) = list(i, link1)
                list(k, link2) = addr(1)
                list(i, link1) = list(k, link2)
                list(j, link2) = i
                CALL adddata
                EXIT loop
              ELSE
                CYCLE loop
              END IF
            ELSE
! Insert 'after' i (depending on search direction).
! Link i into the new element.
              list(i, link2) = addr(1)
            END IF
            CALL addlink
            CALL adddata
            EXIT loop
          END DO loop
        END IF

      CONTAINS
! These two small routines just remove the need to duplicate
! code or have labels and gotos as with the original Algol
        SUBROUTINE addlink
! Insert at one of the ends of the list,
! linking to element i in the opposite direction.
! As a special case, i .EQ. flag when inserting into the empty list.
          j = addr(1)
          list(j, link1) = i
          list(j, link2) = flag
          IF (link2==(m+2)) THEN
            first = addr(1)
          END IF
        END SUBROUTINE addlink

        SUBROUTINE adddata

          list(j, 1) = t
          list(j, 2:m+1) = info(1:m)
          addr(1:n) = [ addr(2:n), 0 ]
        END SUBROUTINE adddata

      END SUBROUTINE inlist

      SUBROUTINE outlist(vector, m, list, n, first, flag, addr)
! outlist removes the first entry (information pair with
! lowest order key) from list(i,j) and puts it in vector(k);
        INTEGER, INTENT (INOUT) :: first
        INTEGER, INTENT (IN) :: flag, m, n
! ..
! .. Array Arguments ..
        INTEGER, INTENT (INOUT) :: addr(:), list(:, :), vector(:)
! ..
! .. Local Scalars ..
        INTEGER :: i
! ..
! Copy the key and data.
        vector(1:m+1) = list(first, 1:m+1)

! Release the first entry.
        i = n
        DO WHILE (addr(i)==0 .AND. i>1)
          i = i - 1
        END DO
        IF (addr(i)==0) THEN
          addr(i) = first
        ELSE
          addr(i+1) = first
        END IF
        IF (list(first,m+3)==flag) THEN
! The list is now empty.
          list(1, m+2) = flag
        ELSE
          first = list(first, m+3)
          list(first, m+2) = flag
        END IF
      END SUBROUTINE outlist
    END MODULE chainlinkmod
