    PROGRAM sorter
      USE chainlinkmod, ONLY: inlist, outlist
      USE set_precision, ONLY: wp

      INTEGER, PARAMETER :: flag = -1, infosize = 3, listsize = 50

! Variables for the parameters.
      INTEGER :: first, m, n, t
      INTEGER, ALLOCATABLE :: addr(:), info(:), list(:, :), vector(:)
      LOGICAL :: listfull

! Variables for the test rig.
      INTEGER :: errno, i, j, k
      INTEGER, ALLOCATABLE :: sorted_keys(:)

      m = infosize
      n = listsize

      WRITE (*, '(a)') 'This test has executed successfully if &
        &this the only line of text output'

      ALLOCATE (addr(n), info(m), list(n,m+3), sorted_keys(n), vector(m+1), &
        STAT=errno)
      IF (errno/=0) THEN
        WRITE (*, '(''Allocate failed trying to generate addr, &
          &info and list arrays'')')
        STOP
      END IF

! Set up the list of free elements.
      addr(1:n) = [ (i,i=1,n) ]

      CALL random_seed()

! Completely fill the list.
      CALL fill_list

! Check that the list is correct to start with.
      CALL check_list(n)

! Gather the list of sorted keys.
      k = first
      DO i = 1, n
        sorted_keys(i) = list(k, 1)
        k = list(k, m+3)
      END DO

! Successively remove elements from the list and check
! that the keys and values are correct after each removal.
      DO i = 1, n
        CALL outlist(vector, m, list, n, first, flag, addr)
        CALL check_free(i)

! Check the key.
        t = vector(1)
        IF (t/=sorted_keys(i)) THEN
          WRITE (*, '(2(a,i6))') 'Key ', t, ' does not match ', sorted_keys(i)
        END IF

! Check the data returned.
        DO j = 1, m
          IF (vector(1+j)/=t+j) THEN
            WRITE (*, '(2(a,i6))') 'info mismatch for key', t, 'element', j
          END IF
        END DO

! Check the free list.
        DO j = 1, i
          IF (addr(j)==0) THEN
            WRITE (*, '(2(a,i6))') 'Free list incorrect at index ', j, &
              'after', i, 'deletions'
          END IF
        END DO
! Check that the list is still sorted correctly.
        CALL check_list(n-i)
      END DO

! Refill the list.
      CALL fill_list

! Cause a list overflow
      CALL inlist(t, info, m, list, n, first, flag, addr, listfull)
      IF (.NOT. listfull) THEN
        WRITE (*, '(a)') 'Test for full list not triggered'
      END IF


      DEALLOCATE (addr, info, list, sorted_keys, vector)

    CONTAINS
      SUBROUTINE fill_list
! Fill the list checking correctly order as we go
! also check the number of free locations is correct
        INTEGER :: i, j, t
        INTEGER, DIMENSION (m) :: info
        REAL (wp) :: r
! Insert n key/value pairs into the list.
        DO i = 1, n
! Set up the key/value pairs.
! Random key.
          CALL random_number(r)
          t = 1 + int(r*1000)/23
! Value to be inserted.
! Make them easy to check against t.
          info(1:m) = [ (t+j,j=1,m) ]
          CALL inlist(t, info, m, list, n, first, flag, addr, listfull)
          CALL check_list(i)
          CALL check_free(n-i)
        END DO

      END SUBROUTINE fill_list

      SUBROUTINE check_free(numfree)
        INTEGER, INTENT(IN) :: numfree
! Check that the addr array has the correct number of free locations
! which should be equal to numfree
        INTEGER :: i, free

        free = 0
        DO i = 1, n
          IF (addr(i) /= 0) THEN
            free = free + 1
          END IF
        END DO

        IF (free /= numfree) THEN
          WRITE(*,ADVANCE='NO',FMT='(a,i6)')'Incorrect number of free locations: ', free
          WRITE(*,'(a,i6)')' which should be: ', numfree
        END IF
      END SUBROUTINE check_free

      SUBROUTINE check_list(list_length)
! Check that list contains list_length keys in ascending order
! starting from list(first).
! The list has length n.
        INTEGER, INTENT (IN) :: list_length

        INTEGER :: checked, i, j, prev_t, t
        LOGICAL :: error

        error = .FALSE.
        IF (first<0 .OR. first>n) THEN
          WRITE (*, '(2(a,i6))') 'first is outside the value range of 1 to ', &
            n, ' = ', first
        ELSE
          checked = 0
! A dummy key for comparison at the start of the list.
          prev_t = -1000
          i = first

          DO WHILE ((.NOT. error) .AND. checked<list_length .AND. i/=flag)
            t = list(i, 1)
            IF (t<prev_t) THEN
              WRITE (*, '(2(a,i6))') t, ' at index ', i, &
                ' is out of order with respect to ', prev_t
              error = .TRUE.
            ELSE
! Check the values associated with this key.
              DO j = 1, m
                IF (list(i,1+j)/=t+j) THEN
                  error = .TRUE.
                  WRITE (*, '(a,2i6)') 'info mismatch for index ', i, j
                END IF
              END DO
            END IF
            checked = checked + 1
! Follow the 'next' link.
            i = list(i, m+3)
          END DO
          IF (checked/=list_length) THEN
            WRITE (*, '(2(a,i6))') 'checking finished after ', checked, &
              ' items rather than ', list_length
          END IF
        END IF
      END SUBROUTINE check_list
    END PROGRAM sorter
